// SPDX-License-Identifier: GPL-2.0-or-later

/* PASST - Plug A Simple Socket Transport
 *  for qemu/UNIX domain socket mode
 *
 * PASTA - Pack A Subtle Tap Abstraction
 *  for network namespace/tap device mode
 *
 * igmp.c - (to be implemented) IGMP proxy
 *
 * Copyright (c) 2020-2021 Red Hat GmbH
 * Author: Stefano Brivio <sbrivio@redhat.com>
 */

/* TO BE IMPLEMENTED */
/* cppcheck-suppress unusedFunction */
__attribute__((__unused__)) static void unused(void) { }
